@@ -60,6 +60,12 @@ impl Builder {
         self
     }
 
+    #[allow(unused)]
+    pub fn with_after_include<S: AsRef<str>>(mut self, line: S) -> Builder {
+        self.config.after_includes = Some(String::from(line.as_ref()));
+        self
+    }
+
     #[allow(unused)]
     pub fn with_trailer<S: AsRef<str>>(mut self, trailer: S) -> Builder {
         self.config.trailer = Some(String::from(trailer.as_ref()));
@@ -72,6 +78,12 @@ impl Builder {
         self
     }
 
+    #[allow(unused)]
+    pub fn with_pragma_once(mut self, pragma_once: bool) -> Builder {
+        self.config.pragma_once = pragma_once;
+        self
+    }
+
     #[allow(unused)]
     pub fn with_autogen_warning<S: AsRef<str>>(mut self, autogen_warning: S) -> Builder {
         self.config.autogen_warning = Some(String::from(autogen_warning.as_ref()));
@@ -301,23 +313,14 @@ impl Builder {
         if let Some((lib_dir, binding_lib_name)) = self.lib.clone() {
             let lockfile = self.lockfile.as_ref().and_then(|p| p.to_str());
 
-            let cargo = if let Some(binding_lib_name) = binding_lib_name {
-                Cargo::load(
-                    &lib_dir,
-                    lockfile,
-                    Some(&binding_lib_name),
-                    self.config.parse.parse_deps,
-                    self.config.parse.clean,
-                )?
-            } else {
-                Cargo::load(
-                    &lib_dir,
-                    lockfile,
-                    None,
-                    self.config.parse.parse_deps,
-                    self.config.parse.clean,
-                )?
-            };
+            let cargo = Cargo::load(
+                &lib_dir,
+                lockfile,
+                binding_lib_name.as_deref(),
+                self.config.parse.parse_deps,
+                self.config.parse.clean,
+                /* existing_metadata = */ None,
+            )?;
 
             result.extend_with(&parser::parse_lib(cargo, &self.config)?);
         } else if let Some(cargo) = self.lib_cargo.clone() {
