@@ -7,14 +7,18 @@ use std::io::Write;
 use crate::bindgen::declarationtyperesolver::DeclarationType;
 use crate::bindgen::ir::{Function, Type};
 use crate::bindgen::writer::{ListType, SourceWriter};
+use crate::bindgen::{Config, Language};
 
 // This code is for translating Rust types into C declarations.
 // See Section 6.7, Declarations, in the C standard for background.
 // http://www.open-std.org/jtc1/sc22/wg14/www/docs/n1570.pdf
 
 enum CDeclarator {
-    Ptr(bool),
-    Ref,
+    Ptr {
+        is_const: bool,
+        is_nullable: bool,
+        is_ref: bool,
+    },
     Array(String),
     Func(Vec<(Option<String>, CDecl)>, bool),
 }
@@ -22,7 +26,7 @@ enum CDeclarator {
 impl CDeclarator {
     fn is_ptr(&self) -> bool {
         match self {
-            CDeclarator::Ptr(..) | CDeclarator::Ref | CDeclarator::Func(..) => true,
+            CDeclarator::Ptr { .. } | CDeclarator::Func(..) => true,
             _ => false,
         }
     }
@@ -62,7 +66,7 @@ impl CDecl {
         let args = f
             .args
             .iter()
-            .map(|&(ref arg_name, ref arg_ty)| (Some(arg_name.clone()), CDecl::from_type(arg_ty)))
+            .map(|arg| (arg.name.clone(), CDecl::from_type(&arg.ty)))
             .collect();
         self.declarators
             .push(CDeclarator::Func(args, layout_vertical));
@@ -112,22 +116,18 @@ impl CDecl {
                 );
                 self.type_name = p.to_string();
             }
-
-            Type::ConstPtr(ref t) => {
-                self.declarators.push(CDeclarator::Ptr(is_const));
-                self.build_type(t, true);
-            }
-            Type::Ptr(ref t) => {
-                self.declarators.push(CDeclarator::Ptr(is_const));
-                self.build_type(t, false);
-            }
-            Type::Ref(ref t) => {
-                self.declarators.push(CDeclarator::Ref);
-                self.build_type(t, true);
-            }
-            Type::MutRef(ref t) => {
-                self.declarators.push(CDeclarator::Ref);
-                self.build_type(t, false);
+            Type::Ptr {
+                ref ty,
+                is_nullable,
+                is_const: ptr_is_const,
+                is_ref,
+            } => {
+                self.declarators.push(CDeclarator::Ptr {
+                    is_const,
+                    is_nullable: *is_nullable,
+                    is_ref: *is_ref,
+                });
+                self.build_type(ty, *ptr_is_const);
             }
             Type::Array(ref t, ref constant) => {
                 let len = constant.as_str().to_owned();
@@ -139,19 +139,18 @@ impl CDecl {
                     .iter()
                     .map(|(ref name, ref ty)| (name.clone(), CDecl::from_type(ty)))
                     .collect();
-                self.declarators.push(CDeclarator::Ptr(false));
+                self.declarators.push(CDeclarator::Ptr {
+                    is_const: false,
+                    is_nullable: true,
+                    is_ref: false,
+                });
                 self.declarators.push(CDeclarator::Func(args, false));
                 self.build_type(ret, false);
             }
         }
     }
 
-    fn write<F: Write>(
-        &self,
-        out: &mut SourceWriter<F>,
-        ident: Option<&str>,
-        void_prototype: bool,
-    ) {
+    fn write<F: Write>(&self, out: &mut SourceWriter<F>, ident: Option<&str>, config: &Config) {
         // Write the type-specifier and type-qualifier first
         if !self.type_qualifers.is_empty() {
             write!(out, "{} ", self.type_qualifers);
@@ -182,15 +181,20 @@ impl CDecl {
             let next_is_pointer = iter_rev.peek().map_or(false, |x| x.is_ptr());
 
             match *declarator {
-                CDeclarator::Ptr(ref is_const) => {
-                    if *is_const {
-                        out.write("*const ");
-                    } else {
-                        out.write("*");
+                CDeclarator::Ptr {
+                    is_const,
+                    is_nullable,
+                    is_ref,
+                } => {
+                    out.write(if is_ref { "&" } else { "*" });
+                    if is_const {
+                        out.write("const ");
+                    }
+                    if !is_nullable && !is_ref {
+                        if let Some(attr) = &config.pointer.non_null_attribute {
+                            write!(out, "{} ", attr);
+                        }
                     }
-                }
-                CDeclarator::Ref => {
-                    out.write("&");
                 }
                 CDeclarator::Array(..) => {
                     if next_is_pointer {
@@ -217,10 +221,7 @@ impl CDecl {
         #[allow(clippy::while_let_on_iterator)]
         while let Some(declarator) = iter.next() {
             match *declarator {
-                CDeclarator::Ptr(..) => {
-                    last_was_pointer = true;
-                }
-                CDeclarator::Ref => {
+                CDeclarator::Ptr { .. } => {
                     last_was_pointer = true;
                 }
                 CDeclarator::Array(ref constant) => {
@@ -237,7 +238,7 @@ impl CDecl {
                     }
 
                     out.write("(");
-                    if args.is_empty() && void_prototype {
+                    if args.is_empty() && config.language == Language::C {
                         out.write("void");
                     }
                     if layout_vertical {
@@ -252,7 +253,7 @@ impl CDecl {
                             // Convert &Option<String> to Option<&str>
                             let arg_ident = arg_ident.as_ref().map(|x| x.as_ref());
 
-                            arg_ty.write(out, arg_ident, void_prototype);
+                            arg_ty.write(out, arg_ident, config);
                         }
                         out.pop_tab();
                     } else {
@@ -264,7 +265,7 @@ impl CDecl {
                             // Convert &Option<String> to Option<&str>
                             let arg_ident = arg_ident.as_ref().map(|x| x.as_ref());
 
-                            arg_ty.write(out, arg_ident, void_prototype);
+                            arg_ty.write(out, arg_ident, config);
                         }
                     }
                     out.write(")");
@@ -280,15 +281,15 @@ pub fn write_func<F: Write>(
     out: &mut SourceWriter<F>,
     f: &Function,
     layout_vertical: bool,
-    void_prototype: bool,
+    config: &Config,
 ) {
-    CDecl::from_func(f, layout_vertical).write(out, Some(f.path().name()), void_prototype);
+    CDecl::from_func(f, layout_vertical).write(out, Some(f.path().name()), config);
 }
 
-pub fn write_field<F: Write>(out: &mut SourceWriter<F>, t: &Type, ident: &str) {
-    CDecl::from_type(t).write(out, Some(ident), false);
+pub fn write_field<F: Write>(out: &mut SourceWriter<F>, t: &Type, ident: &str, config: &Config) {
+    CDecl::from_type(t).write(out, Some(ident), config);
 }
 
-pub fn write_type<F: Write>(out: &mut SourceWriter<F>, t: &Type) {
-    CDecl::from_type(t).write(out, None, false);
+pub fn write_type<F: Write>(out: &mut SourceWriter<F>, t: &Type, config: &Config) {
+    CDecl::from_type(t).write(out, None, config);
 }
