@@ -201,6 +201,28 @@ impl FromStr for ItemType {
 
 deserialize_enum_str!(ItemType);
 
+/// Type which specifies the sort order of functions
+#[derive(Debug, Clone, PartialEq)]
+pub enum SortKey {
+    Name,
+    None,
+}
+
+impl FromStr for SortKey {
+    type Err = String;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        use self::SortKey::*;
+        Ok(match &*s.to_lowercase() {
+            "name" => Name,
+            "none" => None,
+            _ => return Err(format!("Unrecognized sort option: '{}'.", s)),
+        })
+    }
+}
+
+deserialize_enum_str!(SortKey);
+
 /// Settings to apply when exporting items.
 #[derive(Debug, Clone, Deserialize, Default)]
 #[serde(rename_all = "snake_case")]
@@ -214,6 +236,8 @@ pub struct ExportConfig {
     pub exclude: Vec<String>,
     /// Table of name conversions to apply to item names
     pub rename: HashMap<String, String>,
+    /// Table of raw strings to prepend to the body of items.
+    pub pre_body: HashMap<String, String>,
     /// Table of raw strings to append to the body of items.
     pub body: HashMap<String, String>,
     /// A prefix to add before the name of every item
@@ -222,6 +246,8 @@ pub struct ExportConfig {
     pub item_types: Vec<ItemType>,
     /// Whether renaming overrides or extends prefixing.
     pub renaming_overrides_prefixing: bool,
+    /// Name mangling character. Defaults to `_`
+    pub mangle_separator: Option<String>,
 }
 
 impl ExportConfig {
@@ -229,7 +255,11 @@ impl ExportConfig {
         self.item_types.is_empty() || self.item_types.contains(&item_type)
     }
 
-    pub(crate) fn extra_body(&self, path: &Path) -> Option<&str> {
+    pub(crate) fn pre_body(&self, path: &Path) -> Option<&str> {
+        self.pre_body.get(path.name()).map(|s| s.trim_matches('\n'))
+    }
+
+    pub(crate) fn post_body(&self, path: &Path) -> Option<&str> {
         self.body.get(path.name()).map(|s| s.trim_matches('\n'))
     }
 
@@ -279,12 +309,18 @@ pub struct FunctionConfig {
     pub prefix: Option<String>,
     /// Optional text to output after each function declaration
     pub postfix: Option<String>,
-    /// The way to annotation this function as #[must_use].
+    /// The way to annotation this function as #[must_use]
     pub must_use: Option<String>,
     /// The style to layout the args
     pub args: Layout,
     /// The rename rule to apply to function args
     pub rename_args: Option<RenameRule>,
+    /// An optional macro to use when generating Swift function name attributes
+    pub swift_name_macro: Option<String>,
+    /// Sort key for function names
+    pub sort_by: SortKey,
+    /// Optional text to output after functions which return `!`.
+    pub no_return: Option<String>,
 }
 
 impl Default for FunctionConfig {
@@ -295,6 +331,9 @@ impl Default for FunctionConfig {
             must_use: None,
             args: Layout::Auto,
             rename_args: None,
+            swift_name_macro: None,
+            sort_by: SortKey::Name,
+            no_return: None,
         }
     }
 }
@@ -521,12 +560,15 @@ impl EnumConfig {
 pub struct ConstantConfig {
     /// Whether a generated constant can be a static const in C++ mode.
     pub allow_static_const: bool,
+    /// Whether a generated constant should be constexpr in C++ mode.
+    pub allow_constexpr: bool,
 }
 
 impl Default for ConstantConfig {
     fn default() -> ConstantConfig {
         ConstantConfig {
             allow_static_const: true,
+            allow_constexpr: false,
         }
     }
 }
@@ -652,6 +694,16 @@ impl ParseConfig {
     }
 }
 
+/// Settings to apply to pointers
+#[derive(Debug, Clone, Default, Deserialize)]
+#[serde(rename_all = "snake_case")]
+#[serde(deny_unknown_fields)]
+#[serde(default)]
+pub struct PtrConfig {
+    /// Optional attribute to apply to pointers that are required to not be null
+    pub non_null_attribute: Option<String>,
+}
+
 /// A collection of settings to customize the generated bindings.
 #[derive(Debug, Clone, Deserialize)]
 #[serde(rename_all = "snake_case")]
@@ -664,10 +716,14 @@ pub struct Config {
     pub includes: Vec<String>,
     /// A list of additional system includes to put at the beginning of the generated header
     pub sys_includes: Vec<String>,
+    /// Optional verbatim code added after the include blocks
+    pub after_includes: Option<String>,
     /// Optional text to output at the end of the file
     pub trailer: Option<String>,
     /// Optional name to use for an include guard
     pub include_guard: Option<String>,
+    /// Add a `#pragma once` guard
+    pub pragma_once: bool,
     /// Generates no includes at all. Overrides all other include options
     ///
     /// This option is useful when using cbindgen with tools such as python's cffi which
@@ -721,6 +777,9 @@ pub struct Config {
     pub documentation: bool,
     /// How documentation comments should be styled.
     pub documentation_style: DocumentationStyle,
+    /// Configuration options for pointers
+    #[serde(rename = "ptr")]
+    pub pointer: PtrConfig,
 }
 
 impl Default for Config {
@@ -729,8 +788,10 @@ impl Default for Config {
             header: None,
             includes: Vec::new(),
             sys_includes: Vec::new(),
+            after_includes: None,
             trailer: None,
             include_guard: None,
+            pragma_once: false,
             autogen_warning: None,
             include_version: false,
             no_includes: false,
@@ -754,6 +815,7 @@ impl Default for Config {
             defines: HashMap::new(),
             documentation: true,
             documentation_style: DocumentationStyle::Auto,
+            pointer: PtrConfig::default(),
         }
     }
 }
