@@ -17,15 +17,63 @@ use crate::bindgen::library::Library;
 use crate::bindgen::mangle;
 use crate::bindgen::monomorph::Monomorphs;
 use crate::bindgen::rename::{IdentifierType, RenameRule};
+use crate::bindgen::reserved;
 use crate::bindgen::utilities::find_first_some;
 use crate::bindgen::writer::{ListType, Source, SourceWriter};
 
+#[derive(Debug, Clone)]
+pub enum VariantBody {
+    Empty(AnnotationSet),
+    Body {
+        /// The variant field / export name.
+        name: String,
+        /// The struct with all the items.
+        body: Struct,
+    },
+}
+
+impl VariantBody {
+    fn empty() -> Self {
+        Self::Empty(AnnotationSet::new())
+    }
+
+    fn annotations(&self) -> &AnnotationSet {
+        match *self {
+            Self::Empty(ref anno) => anno,
+            Self::Body { ref body, .. } => &body.annotations,
+        }
+    }
+
+    fn is_empty(&self) -> bool {
+        match *self {
+            Self::Empty(..) => true,
+            Self::Body { .. } => false,
+        }
+    }
+
+    fn specialize(
+        &self,
+        generic_values: &[Type],
+        mappings: &[(&Path, &Type)],
+        config: &Config,
+    ) -> Self {
+        match *self {
+            Self::Empty(ref annos) => Self::Empty(annos.clone()),
+            Self::Body { ref name, ref body } => Self::Body {
+                name: name.clone(),
+                body: body.specialize(generic_values, mappings, config),
+            },
+        }
+    }
+}
+
 #[derive(Debug, Clone)]
 pub struct EnumVariant {
     pub name: String,
     pub export_name: String,
     pub discriminant: Option<i64>,
-    pub body: Option<(String, Struct)>,
+    pub body: VariantBody,
+    pub cfg: Option<Cfg>,
     pub documentation: Documentation,
 }
 
@@ -48,11 +96,12 @@ fn value_from_expr(val: &syn::Expr) -> Option<i64> {
 }
 
 impl EnumVariant {
-    pub fn load(
+    fn load(
         is_tagged: bool,
         variant: &syn::Variant,
         generic_params: GenericParams,
         mod_cfg: Option<&Cfg>,
+        self_path: &Path,
     ) -> Result<Self, String> {
         let discriminant = match variant.discriminant {
             Some((_, ref expr)) => match value_from_expr(expr) {
@@ -65,6 +114,7 @@ impl EnumVariant {
         fn parse_fields(
             is_tagged: bool,
             fields: &syn::punctuated::Punctuated<syn::Field, syn::token::Comma>,
+            self_path: &Path,
         ) -> Result<Vec<(String, Type, Documentation)>, String> {
             let mut res = Vec::new();
 
@@ -77,7 +127,8 @@ impl EnumVariant {
             }
 
             for (i, field) in fields.iter().enumerate() {
-                if let Some(ty) = Type::load(&field.ty)? {
+                if let Some(mut ty) = Type::load(&field.ty)? {
+                    ty.replace_self_with(self_path);
                     res.push((
                         match field.ident {
                             Some(ref ident) => ident.to_string(),
@@ -92,54 +143,59 @@ impl EnumVariant {
             Ok(res)
         }
 
+        let variant_cfg = Cfg::append(mod_cfg, Cfg::load(&variant.attrs));
+        let annotations = AnnotationSet::load(&variant.attrs)?;
         let body = match variant.fields {
-            syn::Fields::Unit => None,
+            syn::Fields::Unit => VariantBody::Empty(annotations),
             syn::Fields::Named(ref fields) => {
                 let path = Path::new(format!("{}_Body", variant.ident));
-                Some(Struct::new(
-                    path,
-                    generic_params,
-                    parse_fields(is_tagged, &fields.named)?,
-                    is_tagged,
-                    true,
-                    None,
-                    false,
-                    false,
-                    Cfg::append(mod_cfg, Cfg::load(&variant.attrs)),
-                    AnnotationSet::load(&variant.attrs)?,
-                    Documentation::none(),
-                ))
+                let name = RenameRule::SnakeCase
+                    .apply(&variant.ident.to_string(), IdentifierType::StructMember);
+                VariantBody::Body {
+                    name,
+                    body: Struct::new(
+                        path,
+                        generic_params,
+                        parse_fields(is_tagged, &fields.named, self_path)?,
+                        is_tagged,
+                        true,
+                        None,
+                        false,
+                        false,
+                        None,
+                        annotations,
+                        Documentation::none(),
+                    ),
+                }
             }
             syn::Fields::Unnamed(ref fields) => {
                 let path = Path::new(format!("{}_Body", variant.ident));
-                Some(Struct::new(
-                    path,
-                    generic_params,
-                    parse_fields(is_tagged, &fields.unnamed)?,
-                    is_tagged,
-                    true,
-                    None,
-                    false,
-                    true,
-                    Cfg::append(mod_cfg, Cfg::load(&variant.attrs)),
-                    AnnotationSet::load(&variant.attrs)?,
-                    Documentation::none(),
-                ))
+                let name = RenameRule::SnakeCase
+                    .apply(&variant.ident.to_string(), IdentifierType::StructMember);
+                VariantBody::Body {
+                    name,
+                    body: Struct::new(
+                        path,
+                        generic_params,
+                        parse_fields(is_tagged, &fields.unnamed, self_path)?,
+                        is_tagged,
+                        true,
+                        None,
+                        false,
+                        true,
+                        None,
+                        annotations,
+                        Documentation::none(),
+                    ),
+                }
             }
         };
 
         Ok(EnumVariant::new(
             variant.ident.to_string(),
             discriminant,
-            body.map(|body| {
-                (
-                    RenameRule::SnakeCase.apply_to_pascal_case(
-                        &format!("{}", variant.ident),
-                        IdentifierType::StructMember,
-                    ),
-                    body,
-                )
-            }),
+            body,
+            variant_cfg,
             Documentation::load(&variant.attrs),
         ))
     }
@@ -147,7 +203,8 @@ impl EnumVariant {
     pub fn new(
         name: String,
         discriminant: Option<i64>,
-        body: Option<(String, Struct)>,
+        body: VariantBody,
+        cfg: Option<Cfg>,
         documentation: Documentation,
     ) -> Self {
         let export_name = name.clone();
@@ -156,54 +213,66 @@ impl EnumVariant {
             export_name,
             discriminant,
             body,
+            cfg,
             documentation,
         }
     }
 
     fn add_dependencies(&self, library: &Library, out: &mut Dependencies) {
-        if let Some((_, ref item)) = self.body {
-            item.add_dependencies(library, out);
+        if let VariantBody::Body { ref body, .. } = self.body {
+            body.add_dependencies(library, out);
         }
     }
 
     fn resolve_declaration_types(&mut self, resolver: &DeclarationTypeResolver) {
-        if let Some((_, ref mut ty)) = self.body {
-            ty.resolve_declaration_types(resolver);
+        if let VariantBody::Body { ref mut body, .. } = self.body {
+            body.resolve_declaration_types(resolver);
         }
     }
 
-    fn specialize(&self, generic_values: &[Type], mappings: &[(&Path, &Type)]) -> Self {
+    fn specialize(
+        &self,
+        generic_values: &[Type],
+        mappings: &[(&Path, &Type)],
+        config: &Config,
+    ) -> Self {
         Self::new(
-            mangle::mangle_name(&self.name, generic_values),
+            mangle::mangle_name(
+                &self.name,
+                generic_values,
+                config.export.mangle_separator.as_deref(),
+            ),
             self.discriminant,
-            self.body
-                .as_ref()
-                .map(|&(ref name, ref ty)| (name.clone(), ty.specialize(generic_values, mappings))),
+            self.body.specialize(generic_values, mappings, config),
+            self.cfg.clone(),
             self.documentation.clone(),
         )
     }
 
     fn add_monomorphs(&self, library: &Library, out: &mut Monomorphs) {
-        if let Some((_, ref ty)) = self.body {
-            ty.add_monomorphs(library, out);
+        if let VariantBody::Body { ref body, .. } = self.body {
+            body.add_monomorphs(library, out);
         }
     }
 
     fn mangle_paths(&mut self, monomorphs: &Monomorphs) {
-        if let Some((_, ref mut ty)) = self.body {
-            ty.mangle_paths(monomorphs);
+        if let VariantBody::Body { ref mut body, .. } = self.body {
+            body.mangle_paths(monomorphs);
         }
     }
 }
 
 impl Source for EnumVariant {
     fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
+        let condition = self.cfg.to_condition(config);
+        condition.write_before(config, out);
         self.documentation.write(config, out);
         write!(out, "{}", self.export_name);
         if let Some(discriminant) = self.discriminant {
             write!(out, " = {}", discriminant);
         }
         out.write(",");
+        condition.write_after(config, out);
     }
 }
 
@@ -236,11 +305,9 @@ impl Enum {
             return false;
         }
 
-        self.variants.iter().all(|variant| {
-            variant
-                .body
-                .as_ref()
-                .map_or(true, |&(_, ref body)| body.can_derive_eq())
+        self.variants.iter().all(|variant| match variant.body {
+            VariantBody::Empty(..) => true,
+            VariantBody::Body { ref body, .. } => body.can_derive_eq(),
         })
     }
 
@@ -250,7 +317,11 @@ impl Enum {
         }
     }
 
-    pub fn load(item: &syn::ItemEnum, mod_cfg: Option<&Cfg>) -> Result<Enum, String> {
+    pub fn load(
+        item: &syn::ItemEnum,
+        mod_cfg: Option<&Cfg>,
+        config: &Config,
+    ) -> Result<Enum, String> {
         let repr = Repr::load(&item.attrs)?;
         if repr.style == ReprStyle::Rust && repr.ty.is_none() {
             return Err("Enum is not marked with a valid #[repr(prim)] or #[repr(C)].".to_owned());
@@ -260,6 +331,7 @@ impl Enum {
             return Err("Enum is marked with #[repr(align(...))] or #[repr(packed)].".to_owned());
         }
 
+        let path = Path::new(item.ident.to_string());
         let generic_params = GenericParams::new(&item.generics);
 
         let mut variants = Vec::new();
@@ -271,8 +343,9 @@ impl Enum {
                 variant,
                 generic_params.clone(),
                 mod_cfg,
+                &path,
             )?;
-            is_tagged = is_tagged || variant.body.is_some();
+            is_tagged = is_tagged || !variant.body.is_empty();
             variants.push(variant);
         }
 
@@ -280,16 +353,32 @@ impl Enum {
 
         if let Some(names) = annotations.list("enum-trailing-values") {
             for name in names {
-                variants.push(EnumVariant::new(name, None, None, Documentation::none()));
+                variants.push(EnumVariant::new(
+                    name,
+                    None,
+                    VariantBody::empty(),
+                    None,
+                    Documentation::none(),
+                ));
             }
         }
 
-        let path = Path::new(item.ident.to_string());
+        if config.enumeration.add_sentinel(&annotations) {
+            variants.push(EnumVariant::new(
+                "Sentinel".to_owned(),
+                None,
+                VariantBody::empty(),
+                None,
+                Documentation::simple(" Must be last for serialization purposes"),
+            ));
+        }
+
         let tag = if is_tagged {
             Some("Tag".to_string())
         } else {
             None
         };
+
         Ok(Enum::new(
             path,
             generic_params,
@@ -376,7 +465,7 @@ impl Item for Enum {
             let new_tag = format!("{}_Tag", self.export_name);
             if self.repr.style == ReprStyle::Rust {
                 for variant in &mut self.variants {
-                    if let Some((_, ref mut body)) = variant.body {
+                    if let VariantBody::Body { ref mut body, .. } = variant.body {
                         let path = Path::new(new_tag.clone());
                         let generic_path = GenericPath::new(path, vec![]);
                         body.fields[0].1 = Type::Path(generic_path);
@@ -387,8 +476,15 @@ impl Item for Enum {
         }
 
         for variant in &mut self.variants {
-            if let Some((_, ref mut body)) = variant.body {
+            reserved::escape(&mut variant.export_name);
+
+            if let VariantBody::Body {
+                ref mut name,
+                ref mut body,
+            } = variant.body
+            {
                 body.rename_for_config(config);
+                reserved::escape(name);
             }
         }
 
@@ -397,7 +493,7 @@ impl Item for Enum {
         {
             for variant in &mut self.variants {
                 variant.export_name = format!("{}_{}", self.export_name, variant.export_name);
-                if let Some((_, ref mut body)) = variant.body {
+                if let VariantBody::Body { ref mut body, .. } = variant.body {
                     body.export_name = format!("{}_{}", self.export_name, body.export_name());
                 }
             }
@@ -414,17 +510,16 @@ impl Item for Enum {
                 .iter()
                 .map(|variant| {
                     EnumVariant::new(
-                        r.apply_to_pascal_case(
-                            &variant.export_name,
-                            IdentifierType::EnumVariant(self),
-                        ),
+                        r.apply(&variant.export_name, IdentifierType::EnumVariant(self)),
                         variant.discriminant,
-                        variant.body.as_ref().map(|body| {
-                            (
-                                r.apply_to_snake_case(&body.0, IdentifierType::StructMember),
-                                body.1.clone(),
-                            )
-                        }),
+                        match variant.body {
+                            VariantBody::Empty(..) => variant.body.clone(),
+                            VariantBody::Body { ref name, ref body } => VariantBody::Body {
+                                name: r.apply(&name, IdentifierType::StructMember),
+                                body: body.clone(),
+                            },
+                        },
+                        variant.cfg.clone(),
                         variant.documentation.clone(),
                     )
                 })
@@ -458,19 +553,24 @@ impl Item for Enum {
             .collect::<Vec<_>>();
 
         for variant in &self.variants {
-            if let Some((_, ref body)) = variant.body {
+            if let VariantBody::Body { ref body, .. } = variant.body {
                 body.instantiate_monomorph(generic_values, library, out);
             }
         }
 
-        let mangled_path = mangle::mangle_path(&self.path, generic_values);
+        let mangled_path = mangle::mangle_path(
+            &self.path,
+            generic_values,
+            library.get_config().export.mangle_separator.as_deref(),
+        );
+
         let monomorph = Enum::new(
             mangled_path,
             GenericParams::default(),
             self.repr,
             self.variants
                 .iter()
-                .map(|v| v.specialize(generic_values, &mappings))
+                .map(|v| v.specialize(generic_values, &mappings, library.get_config()))
                 .collect(),
             self.tag.clone(),
             self.cfg.clone(),
@@ -505,8 +605,7 @@ impl Source for Enum {
             ReprType::I8 => "int8_t",
         });
 
-        let condition = (&self.cfg).to_condition(config);
-
+        let condition = self.cfg.to_condition(config);
         condition.write_before(config, out);
 
         self.documentation.write(config, out);
@@ -527,6 +626,14 @@ impl Source for Enum {
 
             write!(out, " {}", self.export_name());
             out.open_brace();
+
+            // Emit the pre_body section, if relevant
+            // Only do this here if we're writing C++, since the struct that wraps everything is starting here.
+            // If we're writing C, we aren't wrapping the enum and variant structs definitions, so the actual enum struct willstart down below
+            if let Some(body) = config.export.pre_body(&self.path) {
+                out.write_raw_block(body);
+                out.new_line();
+            }
         }
 
         let enum_name = if let Some(ref tag) = self.tag {
@@ -583,11 +690,6 @@ impl Source for Enum {
             }
             variant.write(config, out);
         }
-        if config.enumeration.add_sentinel(&self.annotations) {
-            out.new_line();
-            out.new_line();
-            out.write("Sentinel /* this must be last for serialization purposes. */");
-        }
 
         if config.language == Language::C && size.is_none() && config.style.generate_typedef() {
             out.close_brace(false);
@@ -618,11 +720,13 @@ impl Source for Enum {
         if is_tagged {
             // Emit the cases for the structs
             for variant in &self.variants {
-                if let Some((_, ref body)) = variant.body {
+                if let VariantBody::Body { ref body, .. } = variant.body {
                     out.new_line();
                     out.new_line();
-
+                    let condition = variant.cfg.to_condition(config);
+                    condition.write_before(config, out);
                     body.write(config, out);
+                    condition.write_after(config, out);
                 }
             }
 
@@ -642,6 +746,14 @@ impl Source for Enum {
                 }
 
                 out.open_brace();
+
+                // Emit the pre_body section, if relevant
+                // Only do this if we're writing C, since the struct is starting right here.
+                // For C++, the struct wraps all of the above variant structs too, and we write the pre_body section at the begining of that
+                if let Some(body) = config.export.pre_body(&self.path) {
+                    out.write_raw_block(body);
+                    out.new_line();
+                }
             }
 
             // C++ allows accessing only common initial sequence of union
@@ -653,7 +765,8 @@ impl Source for Enum {
                 out.open_brace();
             }
 
-            if config.language == Language::C && !config.style.generate_typedef() {
+            if config.language == Language::C && !size.is_some() && !config.style.generate_typedef()
+            {
                 out.write("enum ");
             }
 
@@ -670,19 +783,26 @@ impl Source for Enum {
                 out.open_brace();
             }
 
-            for (i, &(ref field_name, ref body)) in self
-                .variants
-                .iter()
-                .filter_map(|variant| variant.body.as_ref())
-                .enumerate()
             {
-                if i != 0 {
-                    out.new_line();
-                }
-                if config.style.generate_typedef() {
-                    write!(out, "{} {};", body.export_name(), field_name);
-                } else {
-                    write!(out, "struct {} {};", body.export_name(), field_name);
+                let mut first = true;
+                for variant in &self.variants {
+                    let (field_name, body) = match variant.body {
+                        VariantBody::Body { ref name, ref body } => (name, body),
+                        VariantBody::Empty(..) => continue,
+                    };
+
+                    if !first {
+                        out.new_line();
+                    }
+                    first = false;
+                    let condition = variant.cfg.to_condition(config);
+                    condition.write_before(config, out);
+                    if config.style.generate_typedef() {
+                        write!(out, "{} {};", body.export_name(), field_name);
+                    } else {
+                        write!(out, "struct {} {};", body.export_name(), field_name);
+                    }
+                    condition.write_after(config, out);
                 }
             }
 
@@ -694,30 +814,48 @@ impl Source for Enum {
 
             // Emit convenience methods
             let derive_helper_methods = config.enumeration.derive_helper_methods(&self.annotations);
+            let derive_const_casts = config.enumeration.derive_const_casts(&self.annotations);
+            let derive_mut_casts = config.enumeration.derive_mut_casts(&self.annotations);
             if config.language == Language::Cxx && derive_helper_methods {
                 for variant in &self.variants {
                     out.new_line();
                     out.new_line();
 
+                    let condition = variant.cfg.to_condition(config);
+                    condition.write_before(config, out);
+
                     let arg_renamer = |name: &str| {
                         config
                             .function
                             .rename_args
                             .as_ref()
                             .unwrap_or(&RenameRule::GeckoCase)
-                            .apply_to_snake_case(name, IdentifierType::FunctionArg)
+                            .apply(name, IdentifierType::FunctionArg)
                     };
 
+                    macro_rules! write_attrs {
+                        ($op:expr) => {{
+                            if let Some(Some(attrs)) = variant.body.annotations().atom(concat!(
+                                "variant-",
+                                $op,
+                                "-attributes"
+                            )) {
+                                write!(out, "{} ", attrs);
+                            }
+                        }};
+                    };
+
+                    write_attrs!("constructor");
                     write!(out, "static {} {}(", self.export_name, variant.export_name);
 
-                    if let Some((_, ref body)) = variant.body {
+                    if let VariantBody::Body { ref body, .. } = variant.body {
                         let vec: Vec<_> = body
                             .fields
                             .iter()
                             .skip(skip_fields)
                             .map(|&(ref name, ref ty, _)| {
                                 // const-ref args to constructor
-                                (arg_renamer(name), Type::Ref(Box::new(ty.clone())))
+                                (arg_renamer(name), Type::const_ref_to(ty))
                             })
                             .collect();
                         out.write_vertical_source_list(&vec[..], ListType::Join(","));
@@ -728,7 +866,11 @@ impl Source for Enum {
 
                     write!(out, "{} result;", self.export_name);
 
-                    if let Some((ref variant_name, ref body)) = variant.body {
+                    if let VariantBody::Body {
+                        name: ref variant_name,
+                        ref body,
+                    } = variant.body
+                    {
                         for &(ref field_name, ref ty, ..) in body.fields.iter().skip(skip_fields) {
                             out.new_line();
                             match ty {
@@ -764,54 +906,50 @@ impl Source for Enum {
                     out.new_line();
                     write!(out, "return result;");
                     out.close_brace(false);
-                }
 
-                for variant in &self.variants {
                     out.new_line();
                     out.new_line();
 
+                    write_attrs!("is");
                     // FIXME: create a config for method case
                     write!(out, "bool Is{}() const", variant.export_name);
                     out.open_brace();
                     write!(out, "return tag == {}::{};", enum_name, variant.export_name);
                     out.close_brace(false);
-                }
-            }
 
-            let derive_const_casts = config.enumeration.derive_const_casts(&self.annotations);
-            let derive_mut_casts = config.enumeration.derive_mut_casts(&self.annotations);
-            if config.language == Language::Cxx
-                && derive_helper_methods
-                && (derive_const_casts || derive_mut_casts)
-            {
-                let assert_name = match config.enumeration.cast_assert_name {
-                    Some(ref n) => &**n,
-                    None => "assert",
-                };
-
-                for variant in &self.variants {
-                    let (member_name, body) = match variant.body {
-                        Some((ref member_name, ref body)) => (member_name, body),
-                        None => continue,
+                    let assert_name = match config.enumeration.cast_assert_name {
+                        Some(ref n) => &**n,
+                        None => "assert",
                     };
 
-                    let field_count = body.fields.len() - skip_fields;
-                    if field_count == 0 {
-                        continue;
-                    }
-
-                    let dig = field_count == 1 && body.tuple_struct;
                     let mut derive_casts = |const_casts: bool| {
+                        let (member_name, body) = match variant.body {
+                            VariantBody::Body { ref name, ref body } => (name, body),
+                            VariantBody::Empty(..) => return,
+                        };
+
+                        let field_count = body.fields.len() - skip_fields;
+                        if field_count == 0 {
+                            return;
+                        }
+
                         out.new_line();
                         out.new_line();
 
+                        let dig = field_count == 1 && body.tuple_struct;
+                        if const_casts {
+                            write_attrs!("const-cast");
+                        } else {
+                            write_attrs!("mut-cast");
+                        }
                         if dig {
                             let field = body.fields.iter().skip(skip_fields).next().unwrap();
                             let return_type = field.1.clone();
-                            let return_type = if const_casts {
-                                Type::Ref(Box::new(return_type))
-                            } else {
-                                Type::MutRef(Box::new(return_type))
+                            let return_type = Type::Ptr {
+                                ty: Box::new(return_type),
+                                is_const: const_casts,
+                                is_ref: true,
+                                is_nullable: false,
                             };
                             return_type.write(config, out);
                         } else if const_casts {
@@ -842,21 +980,32 @@ impl Source for Enum {
                     if derive_mut_casts {
                         derive_casts(false)
                     }
+
+                    condition.write_after(config, out);
                 }
             }
 
             let other = if let Some(r) = config.function.rename_args {
-                r.apply_to_snake_case("other", IdentifierType::FunctionArg)
+                r.apply("other", IdentifierType::FunctionArg)
             } else {
                 String::from("other")
             };
 
+            macro_rules! write_attrs {
+                ($op:expr) => {{
+                    if let Some(Some(attrs)) = self.annotations.atom(concat!($op, "-attributes")) {
+                        write!(out, "{} ", attrs);
+                    }
+                }};
+            };
+
             if config.language == Language::Cxx
                 && self.can_derive_eq()
                 && config.structure.derive_eq(&self.annotations)
             {
                 out.new_line();
                 out.new_line();
+                write_attrs!("eq");
                 write!(
                     out,
                     "bool operator==(const {}& {}) const",
@@ -870,8 +1019,15 @@ impl Source for Enum {
                 out.new_line();
                 write!(out, "switch (tag)");
                 out.open_brace();
+                let mut exhaustive = true;
                 for variant in &self.variants {
-                    if let Some((ref variant_name, _)) = variant.body {
+                    if let VariantBody::Body {
+                        name: ref variant_name,
+                        ..
+                    } = variant.body
+                    {
+                        let condition = variant.cfg.to_condition(config);
+                        condition.write_before(config, out);
                         write!(
                             out,
                             "case {}::{}: return {} == {}.{};",
@@ -881,16 +1037,26 @@ impl Source for Enum {
                             other,
                             variant_name
                         );
+                        condition.write_after(config, out);
                         out.new_line();
+                    } else {
+                        exhaustive = false;
                     }
                 }
-                write!(out, "default: return true;");
+                if !exhaustive {
+                    write!(out, "default: break;");
+                }
                 out.close_brace(false);
+
+                out.new_line();
+                write!(out, "return true;");
+
                 out.close_brace(false);
 
                 if config.structure.derive_neq(&self.annotations) {
                     out.new_line();
                     out.new_line();
+                    write_attrs!("neq");
                     write!(
                         out,
                         "bool operator!=(const {}& {}) const",
@@ -926,24 +1092,33 @@ impl Source for Enum {
             {
                 out.new_line();
                 out.new_line();
+                write_attrs!("destructor");
                 write!(out, "~{}()", self.export_name);
                 out.open_brace();
                 write!(out, "switch (tag)");
                 out.open_brace();
+                let mut exhaustive = true;
                 for variant in &self.variants {
-                    if let Some((ref variant_name, ref item)) = variant.body {
+                    if let VariantBody::Body { ref name, ref body } = variant.body {
+                        let condition = variant.cfg.to_condition(config);
+                        condition.write_before(config, out);
                         write!(
                             out,
                             "case {}::{}: {}.~{}(); break;",
                             self.tag.as_ref().unwrap(),
                             variant.export_name,
-                            variant_name,
-                            item.export_name(),
+                            name,
+                            body.export_name(),
                         );
+                        condition.write_after(config, out);
                         out.new_line();
+                    } else {
+                        exhaustive = false;
                     }
                 }
-                write!(out, "default: break;");
+                if !exhaustive {
+                    write!(out, "default: break;");
+                }
                 out.close_brace(false);
                 out.close_brace(false);
             }
@@ -955,6 +1130,7 @@ impl Source for Enum {
             {
                 out.new_line();
                 out.new_line();
+                write_attrs!("copy-constructor");
                 write!(
                     out,
                     "{}(const {}& {})",
@@ -965,22 +1141,30 @@ impl Source for Enum {
                 out.open_brace();
                 write!(out, "switch (tag)");
                 out.open_brace();
+                let mut exhaustive = true;
                 for variant in &self.variants {
-                    if let Some((ref variant_name, ref item)) = variant.body {
+                    if let VariantBody::Body { ref name, ref body } = variant.body {
+                        let condition = variant.cfg.to_condition(config);
+                        condition.write_before(config, out);
                         write!(
                             out,
                             "case {}::{}: ::new (&{}) ({})({}.{}); break;",
                             self.tag.as_ref().unwrap(),
                             variant.export_name,
-                            variant_name,
-                            item.export_name(),
+                            name,
+                            body.export_name(),
                             other,
-                            variant_name,
+                            name,
                         );
+                        condition.write_after(config, out);
                         out.new_line();
+                    } else {
+                        exhaustive = false;
                     }
                 }
-                write!(out, "default: break;");
+                if !exhaustive {
+                    write!(out, "default: break;");
+                }
                 out.close_brace(false);
                 out.close_brace(false);
 
@@ -990,6 +1174,7 @@ impl Source for Enum {
                         .derive_tagged_enum_copy_assignment(&self.annotations)
                 {
                     out.new_line();
+                    write_attrs!("copy-assignment");
                     write!(
                         out,
                         "{}& operator=(const {}& {})",
@@ -1008,7 +1193,9 @@ impl Source for Enum {
                 }
             }
 
-            if let Some(body) = config.export.extra_body(&self.path) {
+            // Emit the post_body section, if relevant
+            if let Some(body) = config.export.post_body(&self.path) {
+                out.new_line();
                 out.write_raw_block(body);
             }
 
@@ -1019,7 +1206,6 @@ impl Source for Enum {
                 out.close_brace(true);
             }
         }
-
         condition.write_after(config, out);
     }
 }
