@@ -7,11 +7,12 @@ use std::io::Write;
 use syn;
 
 use crate::bindgen::cdecl;
-use crate::bindgen::config::{Config, Language, Layout};
+use crate::bindgen::config::{Config, Layout};
 use crate::bindgen::declarationtyperesolver::DeclarationTypeResolver;
 use crate::bindgen::dependencies::Dependencies;
 use crate::bindgen::ir::{
-    AnnotationSet, Cfg, ConditionWrite, Documentation, Path, PrimitiveType, ToCondition, Type,
+    AnnotationSet, Cfg, ConditionWrite, Documentation, GenericPath, Path, PrimitiveType,
+    ToCondition, Type,
 };
 use crate::bindgen::library::Library;
 use crate::bindgen::monomorph::Monomorphs;
@@ -20,84 +21,137 @@ use crate::bindgen::reserved;
 use crate::bindgen::utilities::{find_first_some, IterHelpers};
 use crate::bindgen::writer::{Source, SourceWriter};
 
+#[derive(Debug, Clone)]
+pub struct FunctionArgument {
+    pub name: Option<String>,
+    pub ty: Type,
+}
+
 #[derive(Debug, Clone)]
 pub struct Function {
     pub path: Path,
+    /// Path to the self-type of the function
+    /// If the function is a method, this will contain the path of the type in the impl block
+    pub self_type_path: Option<Path>,
     pub ret: Type,
-    pub args: Vec<(String, Type)>,
+    pub args: Vec<FunctionArgument>,
     pub extern_decl: bool,
     pub cfg: Option<Cfg>,
     pub annotations: AnnotationSet,
     pub documentation: Documentation,
+    pub never_return: bool,
 }
 
 impl Function {
     pub fn load(
         path: Path,
+        self_type_path: Option<&Path>,
         sig: &syn::Signature,
         extern_decl: bool,
         attrs: &[syn::Attribute],
         mod_cfg: Option<&Cfg>,
     ) -> Result<Function, String> {
-        let args = sig.inputs.iter().try_skip_map(|x| x.as_ident_and_type())?;
-        let ret = match sig.output {
+        let mut args = sig.inputs.iter().try_skip_map(|x| x.as_argument())?;
+
+        let mut never_return = false;
+        let mut ret = match sig.output {
             syn::ReturnType::Default => Type::Primitive(PrimitiveType::Void),
             syn::ReturnType::Type(_, ref ty) => {
-                if let Some(x) = Type::load(ty)? {
-                    x
-                } else {
+                if let syn::Type::Never(_) = ty.as_ref() {
+                    never_return = true;
                     Type::Primitive(PrimitiveType::Void)
+                } else {
+                    Type::load(ty)?.unwrap_or_else(|| Type::Primitive(PrimitiveType::Void))
                 }
             }
         };
 
+        if let Some(self_path) = self_type_path {
+            for arg in &mut args {
+                arg.ty.replace_self_with(self_path);
+            }
+            ret.replace_self_with(self_path);
+        }
+
         Ok(Function {
             path,
+            self_type_path: self_type_path.cloned(),
             ret,
             args,
             extern_decl,
             cfg: Cfg::append(mod_cfg, Cfg::load(attrs)),
             annotations: AnnotationSet::load(attrs)?,
             documentation: Documentation::load(attrs),
+            never_return,
         })
     }
 
+    pub fn swift_name(&self) -> Option<String> {
+        // If the symbol name starts with the type name, separate the two components with '.'
+        // so that Swift recognises the association between the method and the type
+        let (ref type_prefix, ref type_name) = match self.self_type_path {
+            Some(ref type_name) => {
+                let type_name = type_name.to_string();
+                if !self.path.name().starts_with(&type_name) {
+                    return Some(self.path.to_string());
+                }
+                (format!("{}.", type_name), type_name)
+            }
+            None => ("".to_string(), "".to_string()),
+        };
+
+        let item_name = self
+            .path
+            .name()
+            .trim_start_matches(type_name)
+            .trim_start_matches('_');
+
+        let item_args = {
+            let mut items = Vec::with_capacity(self.args.len());
+            for arg in self.args.iter() {
+                items.push(format!("{}:", arg.name.as_ref()?.as_str()));
+            }
+            items.join("")
+        };
+        Some(format!("{}{}({})", type_prefix, item_name, item_args))
+    }
+
     pub fn path(&self) -> &Path {
         &self.path
     }
 
     pub fn simplify_standard_types(&mut self) {
         self.ret.simplify_standard_types();
-        for &mut (_, ref mut ty) in &mut self.args {
-            ty.simplify_standard_types();
+        for arg in &mut self.args {
+            arg.ty.simplify_standard_types();
         }
     }
 
     pub fn add_dependencies(&self, library: &Library, out: &mut Dependencies) {
         self.ret.add_dependencies(library, out);
-        for &(_, ref ty) in &self.args {
-            ty.add_dependencies(library, out);
+        for arg in &self.args {
+            arg.ty.add_dependencies(library, out);
         }
     }
 
     pub fn add_monomorphs(&self, library: &Library, out: &mut Monomorphs) {
         self.ret.add_monomorphs(library, out);
-        for &(_, ref ty) in &self.args {
-            ty.add_monomorphs(library, out);
+        for arg in &self.args {
+            arg.ty.add_monomorphs(library, out);
         }
     }
 
     pub fn mangle_paths(&mut self, monomorphs: &Monomorphs) {
         self.ret.mangle_paths(monomorphs);
-        for &mut (_, ref mut ty) in &mut self.args {
-            ty.mangle_paths(monomorphs);
+        for arg in &mut self.args {
+            arg.ty.mangle_paths(monomorphs);
         }
     }
 
     pub fn resolve_declaration_types(&mut self, resolver: &DeclarationTypeResolver) {
         self.ret.resolve_declaration_types(resolver);
-        for &mut (_, ref mut ty) in &mut self.args {
-            ty.resolve_declaration_types(resolver);
+        for arg in &mut self.args {
+            arg.ty.resolve_declaration_types(resolver);
         }
     }
 
@@ -105,9 +159,6 @@ impl Function {
         // Rename the types used in arguments
         let generic_params = Default::default();
         self.ret.rename_for_config(config, &generic_params);
-        for &mut (_, ref mut ty) in &mut self.args {
-            ty.rename_for_config(config, &generic_params);
-        }
 
         // Apply rename rules to argument names
         let rules = [
@@ -116,21 +167,23 @@ impl Function {
         ];
 
         if let Some(r) = find_first_some(&rules) {
-            self.args = self
-                .args
-                .iter()
-                .map(|x| {
-                    (
-                        r.apply_to_snake_case(&x.0, IdentifierType::FunctionArg),
-                        x.1.clone(),
-                    )
+            let args = std::mem::replace(&mut self.args, vec![]);
+            self.args = args
+                .into_iter()
+                .map(|arg| {
+                    let name = arg.name.map(|n| r.apply(&n, IdentifierType::FunctionArg));
+                    FunctionArgument { name, ty: arg.ty }
                 })
                 .collect()
         }
 
-        // Escape C/C++ reserved keywords used in argument names
-        for args in &mut self.args {
-            reserved::escape(&mut args.0);
+        // Escape C/C++ reserved keywords used in argument names, and
+        // recursively rename argument types.
+        for arg in &mut self.args {
+            arg.ty.rename_for_config(config, &generic_params);
+            if let Some(ref mut name) = arg.name {
+                reserved::escape(name);
+            }
         }
     }
 }
@@ -138,11 +191,10 @@ impl Function {
 impl Source for Function {
     fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
         fn write_1<W: Write>(func: &Function, config: &Config, out: &mut SourceWriter<W>) {
-            let void_prototype = config.language == Language::C;
             let prefix = config.function.prefix(&func.annotations);
             let postfix = config.function.postfix(&func.annotations);
 
-            let condition = (&func.cfg).to_condition(config);
+            let condition = func.cfg.to_condition(config);
             condition.write_before(config, out);
 
             func.documentation.write(config, out);
@@ -159,24 +211,36 @@ impl Source for Function {
                     }
                 }
             }
-            cdecl::write_func(out, &func, false, void_prototype);
+            cdecl::write_func(out, &func, false, config);
+
             if !func.extern_decl {
                 if let Some(ref postfix) = postfix {
-                    out.write(" ");
-                    write!(out, "{}", postfix);
+                    write!(out, " {}", postfix);
                 }
             }
+
+            if let Some(ref swift_name_macro) = config.function.swift_name_macro {
+                if let Some(swift_name) = func.swift_name() {
+                    write!(out, " {}({})", swift_name_macro, swift_name);
+                }
+            }
+
+            if func.never_return {
+                if let Some(ref no_return_attr) = config.function.no_return {
+                    out.write_fmt(format_args!(" {}", no_return_attr));
+                }
+            }
+
             out.write(";");
 
             condition.write_after(config, out);
         }
 
         fn write_2<W: Write>(func: &Function, config: &Config, out: &mut SourceWriter<W>) {
-            let void_prototype = config.language == Language::C;
             let prefix = config.function.prefix(&func.annotations);
             let postfix = config.function.postfix(&func.annotations);
 
-            let condition = (&func.cfg).to_condition(config);
+            let condition = func.cfg.to_condition(config);
 
             condition.write_before(config, out);
 
@@ -196,13 +260,26 @@ impl Source for Function {
                     }
                 }
             }
-            cdecl::write_func(out, &func, true, void_prototype);
+            cdecl::write_func(out, &func, true, config);
             if !func.extern_decl {
                 if let Some(ref postfix) = postfix {
                     out.new_line();
                     write!(out, "{}", postfix);
                 }
             }
+
+            if let Some(ref swift_name_macro) = config.function.swift_name_macro {
+                if let Some(swift_name) = func.swift_name() {
+                    write!(out, " {}({})", swift_name_macro, swift_name);
+                }
+            }
+
+            if func.never_return {
+                if let Some(ref no_return_attr) = config.function.no_return {
+                    out.write_fmt(format_args!(" {}", no_return_attr));
+                }
+            }
+
             out.write(";");
 
             condition.write_after(config, out);
@@ -220,26 +297,54 @@ impl Source for Function {
     }
 }
 
-pub trait SynFnArgHelpers {
-    fn as_ident_and_type(&self) -> Result<Option<(String, Type)>, String>;
+trait SynFnArgHelpers {
+    fn as_argument(&self) -> Result<Option<FunctionArgument>, String>;
+}
+
+fn gen_self_type(receiver: &syn::Receiver) -> Type {
+    let self_ty = Type::Path(GenericPath::self_path());
+    if receiver.reference.is_none() {
+        return self_ty;
+    }
+
+    let is_const = receiver.mutability.is_none();
+    Type::Ptr {
+        ty: Box::new(self_ty),
+        is_const,
+        is_nullable: false,
+        is_ref: false,
+    }
 }
 
 impl SynFnArgHelpers for syn::FnArg {
-    fn as_ident_and_type(&self) -> Result<Option<(String, Type)>, String> {
-        match self {
-            &syn::FnArg::Typed(syn::PatType {
+    fn as_argument(&self) -> Result<Option<FunctionArgument>, String> {
+        match *self {
+            syn::FnArg::Typed(syn::PatType {
                 ref pat, ref ty, ..
-            }) => match **pat {
-                syn::Pat::Ident(syn::PatIdent { ref ident, .. }) => {
-                    if let Some(x) = Type::load(ty)? {
-                        Ok(Some((ident.to_string(), x)))
-                    } else {
-                        Ok(None)
+            }) => {
+                let name = match **pat {
+                    syn::Pat::Wild(..) => None,
+                    syn::Pat::Ident(syn::PatIdent { ref ident, .. }) => Some(ident.to_string()),
+                    _ => {
+                        return Err(format!(
+                            "Parameter has an unsupported argument name: {:?}",
+                            pat
+                        ))
                     }
+                };
+                let ty = match Type::load(ty)? {
+                    Some(x) => x,
+                    None => return Ok(None),
+                };
+                if let Type::Array(..) = ty {
+                    return Err("Array as function arguments are not supported".to_owned());
                 }
-                _ => Err("Parameter has an unsupported type.".to_owned()),
-            },
-            _ => Err("Parameter has an unsupported type.".to_owned()),
+                Ok(Some(FunctionArgument { name, ty }))
+            }
+            syn::FnArg::Receiver(ref receiver) => Ok(Some(FunctionArgument {
+                name: Some("self".to_string()),
+                ty: gen_self_type(receiver),
+            })),
         }
     }
 }
