@@ -27,6 +27,32 @@ impl GenericParams {
                 .collect(),
         )
     }
+
+    fn write_internal<F: Write>(
+        &self,
+        config: &Config,
+        out: &mut SourceWriter<F>,
+        with_default: bool,
+    ) {
+        if !self.0.is_empty() && config.language == Language::Cxx {
+            out.write("template<");
+            for (i, item) in self.0.iter().enumerate() {
+                if i != 0 {
+                    out.write(", ");
+                }
+                write!(out, "typename {}", item);
+                if with_default {
+                    write!(out, " = void");
+                }
+            }
+            out.write(">");
+            out.new_line();
+        }
+    }
+
+    pub fn write_with_default<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
+        self.write_internal(config, out, true);
+    }
 }
 
 impl Deref for GenericParams {
@@ -39,17 +65,7 @@ impl Deref for GenericParams {
 
 impl Source for GenericParams {
     fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
-        if !self.0.is_empty() && config.language == Language::Cxx {
-            out.write("template<");
-            for (i, item) in self.0.iter().enumerate() {
-                if i != 0 {
-                    out.write(", ");
-                }
-                write!(out, "typename {}", item);
-            }
-            out.write(">");
-            out.new_line();
-        }
+        self.write_internal(config, out, false);
     }
 }
 
@@ -72,6 +88,10 @@ impl GenericPath {
         }
     }
 
+    pub fn self_path() -> Self {
+        Self::new(Path::new("Self"), vec![])
+    }
+
     pub fn replace_self_with(&mut self, self_ty: &Path) {
         if self.path.replace_self_with(self_ty) {
             self.export_name = self_ty.name().to_owned();
