@@ -65,6 +65,8 @@ impl Struct {
             }
         };
 
+        let path = Path::new(item.ident.to_string());
+
         // Ensure we can safely represent the struct given the configuration.
         if let Some(align) = repr.align {
             layout_config.ensure_safe_to_represent(&align)?;
@@ -76,14 +78,15 @@ impl Struct {
                 let out = fields
                     .named
                     .iter()
-                    .try_skip_map(|x| x.as_ident_and_type())?;
+                    .try_skip_map(|x| x.as_ident_and_type(&path))?;
                 (out, false)
             }
             syn::Fields::Unnamed(ref fields) => {
                 let mut out = Vec::new();
                 let mut current = 0;
                 for field in fields.unnamed.iter() {
-                    if let Some(x) = Type::load(&field.ty)? {
+                    if let Some(mut x) = Type::load(&field.ty)? {
+                        x.replace_self_with(&path);
                         out.push((format!("{}", current), x, Documentation::load(&field.attrs)));
                         current += 1;
                     }
@@ -96,7 +99,7 @@ impl Struct {
         let is_enum_variant_body = false;
 
         Ok(Struct::new(
-            Path::new(item.ident.to_string()),
+            path,
             GenericParams::new(&item.generics),
             fields,
             is_tagged,
@@ -169,8 +172,17 @@ impl Struct {
         }
     }
 
-    pub fn specialize(&self, generic_values: &[Type], mappings: &[(&Path, &Type)]) -> Self {
-        let mangled_path = mangle::mangle_path(&self.path, generic_values);
+    pub fn specialize(
+        &self,
+        generic_values: &[Type],
+        mappings: &[(&Path, &Type)],
+        config: &Config,
+    ) -> Self {
+        let mangled_path = mangle::mangle_path(
+            &self.path,
+            generic_values,
+            config.export.mangle_separator.as_deref(),
+        );
         Struct::new(
             mangled_path,
             GenericParams::default(),
@@ -307,7 +319,7 @@ impl Item for Struct {
                 }
             } else if let Some(r) = find_first_some(&field_rules) {
                 for name in names {
-                    *name = r.apply_to_snake_case(name, IdentifierType::StructMember);
+                    *name = r.apply(name, IdentifierType::StructMember);
                 }
             } else if self.tuple_struct {
                 // If there is a tag field, skip it
@@ -374,7 +386,7 @@ impl Item for Struct {
             .zip(generic_values.iter())
             .collect::<Vec<_>>();
 
-        let monomorph = self.specialize(generic_values, &mappings);
+        let monomorph = self.specialize(generic_values, &mappings, library.get_config());
 
         // Instantiate any monomorphs for any generic paths we may have just created.
         monomorph.add_monomorphs(library, out);
@@ -403,7 +415,7 @@ impl Source for Struct {
             return;
         }
 
-        let condition = (&self.cfg).to_condition(config);
+        let condition = self.cfg.to_condition(config);
         condition.write_before(config, out);
 
         self.documentation.write(config, out);
@@ -452,6 +464,12 @@ impl Source for Struct {
 
         out.open_brace();
 
+        // Emit the pre_body section, if relevant
+        if let Some(body) = config.export.pre_body(&self.path) {
+            out.write_raw_block(body);
+            out.new_line();
+        }
+
         if config.documentation {
             out.write_vertical_source_list(&self.fields, ListType::Cap(";"));
         } else {
@@ -480,7 +498,7 @@ impl Source for Struct {
                         .rename_args
                         .as_ref()
                         .unwrap_or(&RenameRule::GeckoCase)
-                        .apply_to_snake_case(name, IdentifierType::FunctionArg)
+                        .apply(name, IdentifierType::FunctionArg)
                 };
                 write!(out, "{}(", self.export_name());
                 let vec: Vec<_> = self
@@ -507,7 +525,7 @@ impl Source for Struct {
             }
 
             let other = if let Some(r) = config.function.rename_args {
-                r.apply_to_snake_case("other", IdentifierType::FunctionArg)
+                r.apply("other", IdentifierType::FunctionArg)
             } else {
                 String::from("other")
             };
@@ -540,67 +558,83 @@ impl Source for Struct {
 
             let skip_fields = if self.is_tagged { 1 } else { 0 };
 
-            let mut emit_op = |op, conjuc| {
-                if !wrote_start_newline {
-                    wrote_start_newline = true;
+            macro_rules! emit_op {
+                ($op_name:expr, $op:expr, $conjuc:expr) => {{
+                    if !wrote_start_newline {
+                        #[allow(unused_assignments)]
+                        {
+                            wrote_start_newline = true;
+                        }
+                        out.new_line();
+                    }
+
                     out.new_line();
-                }
 
-                out.new_line();
+                    if let Some(Some(attrs)) =
+                        self.annotations.atom(concat!($op_name, "-attributes"))
+                    {
+                        write!(out, "{} ", attrs);
+                    }
 
-                write!(
-                    out,
-                    "bool operator{}(const {}& {}) const",
-                    op,
-                    self.export_name(),
-                    other
-                );
-                out.open_brace();
-                out.write("return ");
-                let vec: Vec<_> = self
-                    .fields
-                    .iter()
-                    .skip(skip_fields)
-                    .map(|x| format!("{} {} {}.{}", x.0, op, other, x.0))
-                    .collect();
-                out.write_vertical_source_list(&vec[..], ListType::Join(&format!(" {}", conjuc)));
-                out.write(";");
-                out.close_brace(false);
+                    write!(
+                        out,
+                        "bool operator{}(const {}& {}) const",
+                        $op,
+                        self.export_name(),
+                        other
+                    );
+                    out.open_brace();
+                    out.write("return ");
+                    let vec: Vec<_> = self
+                        .fields
+                        .iter()
+                        .skip(skip_fields)
+                        .map(|x| format!("{} {} {}.{}", x.0, $op, other, x.0))
+                        .collect();
+                    out.write_vertical_source_list(
+                        &vec[..],
+                        ListType::Join(&format!(" {}", $conjuc)),
+                    );
+                    out.write(";");
+                    out.close_brace(false);
+                }};
             };
 
             if config.structure.derive_eq(&self.annotations) && self.can_derive_eq() {
-                emit_op("==", "&&");
+                emit_op!("eq", "==", "&&");
             }
             if config.structure.derive_neq(&self.annotations) && self.can_derive_eq() {
-                emit_op("!=", "||");
+                emit_op!("neq", "!=", "||");
             }
             if config.structure.derive_lt(&self.annotations)
                 && self.fields.len() == 1
                 && self.fields[0].1.can_cmp_order()
             {
-                emit_op("<", "&&");
+                emit_op!("lt", "<", "&&");
             }
             if config.structure.derive_lte(&self.annotations)
                 && self.fields.len() == 1
                 && self.fields[0].1.can_cmp_order()
             {
-                emit_op("<=", "&&");
+                emit_op!("lte", "<=", "&&");
             }
             if config.structure.derive_gt(&self.annotations)
                 && self.fields.len() == 1
                 && self.fields[0].1.can_cmp_order()
             {
-                emit_op(">", "&&");
+                emit_op!("gt", ">", "&&");
             }
             if config.structure.derive_gte(&self.annotations)
                 && self.fields.len() == 1
                 && self.fields[0].1.can_cmp_order()
             {
-                emit_op(">=", "&&");
+                emit_op!("gte", ">=", "&&");
             }
         }
 
-        if let Some(body) = config.export.extra_body(&self.path) {
+        // Emit the post_body section, if relevant
+        if let Some(body) = config.export.post_body(&self.path) {
+            out.new_line();
             out.write_raw_block(body);
         }
 
@@ -631,11 +665,17 @@ impl Source for Struct {
 }
 
 pub trait SynFieldHelpers {
-    fn as_ident_and_type(&self) -> Result<Option<(String, Type, Documentation)>, String>;
+    fn as_ident_and_type(
+        &self,
+        self_path: &Path,
+    ) -> Result<Option<(String, Type, Documentation)>, String>;
 }
 
 impl SynFieldHelpers for syn::Field {
-    fn as_ident_and_type(&self) -> Result<Option<(String, Type, Documentation)>, String> {
+    fn as_ident_and_type(
+        &self,
+        self_path: &Path,
+    ) -> Result<Option<(String, Type, Documentation)>, String> {
         let ident = self
             .ident
             .as_ref()
@@ -643,7 +683,8 @@ impl SynFieldHelpers for syn::Field {
             .clone();
         let converted_ty = Type::load(&self.ty)?;
 
-        if let Some(x) = converted_ty {
+        if let Some(mut x) = converted_ty {
+            x.replace_self_with(self_path);
             Ok(Some((
                 ident.to_string(),
                 x,
