@@ -209,10 +209,15 @@ impl ArrayLength {
 
 #[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub enum Type {
-    ConstPtr(Box<Type>),
-    Ptr(Box<Type>),
-    Ref(Box<Type>),
-    MutRef(Box<Type>),
+    Ptr {
+        ty: Box<Type>,
+        is_const: bool,
+        is_nullable: bool,
+        // FIXME: This is a bit of a hack, this is only to get us to codegen
+        // `T&` / `const T&`, but we should probably pass that down as an option
+        // to code generation or something.
+        is_ref: bool,
+    },
     Path(GenericPath),
     Primitive(PrimitiveType),
     Array(Box<Type>, ArrayLength),
@@ -220,6 +225,15 @@ pub enum Type {
 }
 
 impl Type {
+    pub fn const_ref_to(ty: &Self) -> Self {
+        Type::Ptr {
+            ty: Box::new(ty.clone()),
+            is_const: true,
+            is_nullable: false,
+            is_ref: true,
+        }
+    }
+
     pub fn load(ty: &syn::Type) -> Result<Option<Type>, String> {
         let converted = match *ty {
             syn::Type::Reference(ref reference) => {
@@ -234,9 +248,13 @@ impl Type {
                     }
                 };
 
-                match reference.mutability {
-                    Some(_) => Type::Ptr(Box::new(converted)),
-                    None => Type::ConstPtr(Box::new(converted)),
+                // TODO(emilio): we could make these use is_ref: true.
+                let is_const = reference.mutability.is_none();
+                Type::Ptr {
+                    ty: Box::new(converted),
+                    is_const,
+                    is_nullable: false,
+                    is_ref: false,
                 }
             }
             syn::Type::Ptr(ref pointer) => {
@@ -251,9 +269,12 @@ impl Type {
                     }
                 };
 
-                match pointer.mutability {
-                    Some(_) => Type::Ptr(Box::new(converted)),
-                    None => Type::ConstPtr(Box::new(converted)),
+                let is_const = pointer.mutability.is_none();
+                Type::Ptr {
+                    ty: Box::new(converted),
+                    is_const,
+                    is_nullable: true,
+                    is_ref: false,
                 }
             }
             syn::Type::Path(ref path) => {
@@ -358,7 +379,7 @@ impl Type {
     pub fn is_primitive_or_ptr_primitive(&self) -> bool {
         match *self {
             Type::Primitive(..) => true,
-            Type::ConstPtr(ref x) => match x.as_ref() {
+            Type::Ptr { ref ty, .. } => match ty.as_ref() {
                 Type::Primitive(..) => true,
                 _ => false,
             },
@@ -368,13 +389,30 @@ impl Type {
 
     pub fn is_repr_ptr(&self) -> bool {
         match *self {
-            Type::Ptr(..) => true,
-            Type::ConstPtr(..) => true,
+            Type::Ptr { .. } => true,
             Type::FuncPtr(..) => true,
             _ => false,
         }
     }
 
+    pub fn make_nullable(&self) -> Option<Self> {
+        match self.clone() {
+            Type::Ptr {
+                ty,
+                is_const,
+                is_ref,
+                ..
+            } => Some(Type::Ptr {
+                ty,
+                is_const,
+                is_ref,
+                is_nullable: true,
+            }),
+            Type::FuncPtr(x, y) => Some(Type::FuncPtr(x, y)),
+            _ => None,
+        }
+    }
+
     fn simplified_type(&self) -> Option<Self> {
         let path = match *self {
             Type::Path(ref p) => p,
@@ -390,8 +428,14 @@ impl Type {
 
         match path.name() {
             // FIXME(#223): This is not quite correct.
-            "Option" if generic.is_repr_ptr() => Some(generic),
-            "NonNull" => Some(Type::Ptr(Box::new(generic))),
+            "Option" if generic.is_repr_ptr() => generic.make_nullable(),
+            "NonNull" => Some(Type::Ptr {
+                ty: Box::new(generic),
+                is_const: false,
+                is_nullable: false,
+                is_ref: false,
+            }),
+            "Cell" => Some(generic),
             _ => None,
         }
     }
@@ -404,11 +448,9 @@ impl Type {
 
     pub fn replace_self_with(&mut self, self_ty: &Path) {
         match *self {
-            Type::Array(ref mut ty, ..)
-            | Type::MutRef(ref mut ty)
-            | Type::Ref(ref mut ty)
-            | Type::Ptr(ref mut ty)
-            | Type::ConstPtr(ref mut ty) => ty.replace_self_with(self_ty),
+            Type::Array(ref mut ty, ..) | Type::Ptr { ref mut ty, .. } => {
+                ty.replace_self_with(self_ty)
+            }
             Type::Path(ref mut generic_path) => {
                 generic_path.replace_self_with(self_ty);
             }
@@ -426,10 +468,7 @@ impl Type {
         let mut current = self;
         loop {
             match *current {
-                Type::ConstPtr(ref ty) => current = ty,
-                Type::Ptr(ref ty) => current = ty,
-                Type::Ref(ref ty) => current = ty,
-                Type::MutRef(ref ty) => current = ty,
+                Type::Ptr { ref ty, .. } => current = ty,
                 Type::Path(ref generic) => {
                     return Some(generic.path().clone());
                 }
@@ -448,10 +487,17 @@ impl Type {
 
     pub fn specialize(&self, mappings: &[(&Path, &Type)]) -> Type {
         match *self {
-            Type::ConstPtr(ref ty) => Type::ConstPtr(Box::new(ty.specialize(mappings))),
-            Type::Ptr(ref ty) => Type::Ptr(Box::new(ty.specialize(mappings))),
-            Type::Ref(ref ty) => Type::Ref(Box::new(ty.specialize(mappings))),
-            Type::MutRef(ref ty) => Type::MutRef(Box::new(ty.specialize(mappings))),
+            Type::Ptr {
+                ref ty,
+                is_const,
+                is_nullable,
+                is_ref,
+            } => Type::Ptr {
+                ty: Box::new(ty.specialize(mappings)),
+                is_const,
+                is_nullable,
+                is_ref,
+            },
             Type::Path(ref generic_path) => {
                 for &(param, value) in mappings {
                     if generic_path.path() == param {
@@ -490,13 +536,7 @@ impl Type {
         out: &mut Dependencies,
     ) {
         match *self {
-            Type::ConstPtr(ref ty) => {
-                ty.add_dependencies_ignoring_generics(generic_params, library, out);
-            }
-            Type::Ptr(ref ty) => {
-                ty.add_dependencies_ignoring_generics(generic_params, library, out);
-            }
-            Type::Ref(ref ty) | Type::MutRef(ref ty) => {
+            Type::Ptr { ref ty, .. } => {
                 ty.add_dependencies_ignoring_generics(generic_params, library, out);
             }
             Type::Path(ref generic) => {
@@ -544,13 +584,7 @@ impl Type {
 
     pub fn add_monomorphs(&self, library: &Library, out: &mut Monomorphs) {
         match *self {
-            Type::ConstPtr(ref ty) => {
-                ty.add_monomorphs(library, out);
-            }
-            Type::Ptr(ref ty) => {
-                ty.add_monomorphs(library, out);
-            }
-            Type::Ref(ref ty) | Type::MutRef(ref ty) => {
+            Type::Ptr { ref ty, .. } => {
                 ty.add_monomorphs(library, out);
             }
             Type::Path(ref generic) => {
@@ -580,13 +614,7 @@ impl Type {
 
     pub fn rename_for_config(&mut self, config: &Config, generic_params: &GenericParams) {
         match *self {
-            Type::ConstPtr(ref mut ty) => {
-                ty.rename_for_config(config, generic_params);
-            }
-            Type::Ptr(ref mut ty) => {
-                ty.rename_for_config(config, generic_params);
-            }
-            Type::Ref(ref mut ty) | Type::MutRef(ref mut ty) => {
+            Type::Ptr { ref mut ty, .. } => {
                 ty.rename_for_config(config, generic_params);
             }
             Type::Path(ref mut ty) => {
@@ -608,13 +636,7 @@ impl Type {
 
     pub fn resolve_declaration_types(&mut self, resolver: &DeclarationTypeResolver) {
         match *self {
-            Type::ConstPtr(ref mut ty) => {
-                ty.resolve_declaration_types(resolver);
-            }
-            Type::Ptr(ref mut ty) => {
-                ty.resolve_declaration_types(resolver);
-            }
-            Type::Ref(ref mut ty) | Type::MutRef(ref mut ty) => {
+            Type::Ptr { ref mut ty, .. } => {
                 ty.resolve_declaration_types(resolver);
             }
             Type::Path(ref mut generic_path) => {
@@ -635,13 +657,7 @@ impl Type {
 
     pub fn mangle_paths(&mut self, monomorphs: &Monomorphs) {
         match *self {
-            Type::ConstPtr(ref mut ty) => {
-                ty.mangle_paths(monomorphs);
-            }
-            Type::Ptr(ref mut ty) => {
-                ty.mangle_paths(monomorphs);
-            }
-            Type::Ref(ref mut ty) | Type::MutRef(ref mut ty) => {
+            Type::Ptr { ref mut ty, .. } => {
                 ty.mangle_paths(monomorphs);
             }
             Type::Path(ref mut generic_path) => {
@@ -652,7 +668,7 @@ impl Type {
                 if let Some(mangled_path) = monomorphs.mangle_path(&generic_path) {
                     *generic_path = GenericPath::new(mangled_path.clone(), vec![]);
                 } else {
-                    error!(
+                    warn!(
                         "Cannot find a mangling for generic path {:?}. This usually means that a \
                          type referenced by this generic was incompatible or not found.",
                         generic_path
@@ -674,9 +690,7 @@ impl Type {
 
     pub fn can_cmp_order(&self) -> bool {
         match *self {
-            Type::ConstPtr(..) => true,
-            Type::Ptr(..) => true,
-            Type::Ref(..) | Type::MutRef(..) => false,
+            Type::Ptr { is_ref, .. } => !is_ref,
             Type::Path(..) => true,
             Type::Primitive(ref p) => p.can_cmp_order(),
             Type::Array(..) => false,
@@ -686,9 +700,7 @@ impl Type {
 
     pub fn can_cmp_eq(&self) -> bool {
         match *self {
-            Type::ConstPtr(..) => true,
-            Type::Ptr(..) => true,
-            Type::Ref(..) | Type::MutRef(..) => false,
+            Type::Ptr { ref ty, is_ref, .. } => !is_ref || ty.can_cmp_eq(),
             Type::Path(..) => true,
             Type::Primitive(ref p) => p.can_cmp_eq(),
             Type::Array(..) => false,
@@ -704,20 +716,20 @@ impl Source for String {
 }
 
 impl Source for Type {
-    fn write<F: Write>(&self, _config: &Config, out: &mut SourceWriter<F>) {
-        cdecl::write_type(out, &self);
+    fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
+        cdecl::write_type(out, self, config);
     }
 }
 
 impl Source for (String, Type) {
-    fn write<F: Write>(&self, _config: &Config, out: &mut SourceWriter<F>) {
-        cdecl::write_field(out, &self.1, &self.0);
+    fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
+        cdecl::write_field(out, &self.1, &self.0, config);
     }
 }
 
 impl Source for (String, Type, Documentation) {
     fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
         self.2.write(config, out);
-        cdecl::write_field(out, &self.1, &self.0);
+        cdecl::write_field(out, &self.1, &self.0, config);
     }
 }
