@@ -6,7 +6,7 @@ use std::collections::HashMap;
 use std::mem;
 
 use crate::bindgen::bindings::Bindings;
-use crate::bindgen::config::{Config, Language};
+use crate::bindgen::config::{Config, Language, SortKey};
 use crate::bindgen::declarationtyperesolver::DeclarationTypeResolver;
 use crate::bindgen::dependencies::Dependencies;
 use crate::bindgen::error::Error;
@@ -54,14 +54,22 @@ impl Library {
     }
 
     pub fn generate(mut self) -> Result<Bindings, Error> {
-        self.remove_excluded();
-        self.functions.sort_by(|x, y| x.path.cmp(&y.path));
         self.transfer_annotations();
         self.simplify_standard_types();
 
+        match self.config.function.sort_by {
+            SortKey::Name => {
+                self.functions.sort_by(|x, y| x.path.cmp(&y.path));
+            }
+            SortKey::None => { /* keep input order */ }
+        }
+
         if self.config.language == Language::C {
             self.instantiate_monomorphs();
+            self.remove_excluded();
             self.resolve_declaration_types();
+        } else {
+            self.remove_excluded();
         }
 
         self.rename_items();
@@ -80,9 +88,7 @@ impl Library {
         for name in &self.config.export.include {
             let path = Path::new(name.clone());
             if let Some(items) = self.get_items(&path) {
-                if !dependencies.items.contains(&path) {
-                    dependencies.items.insert(path);
-
+                if dependencies.items.insert(path) {
                     for item in &items {
                         item.deref().add_dependencies(&self, &mut dependencies);
                     }
@@ -143,6 +149,10 @@ impl Library {
         None
     }
 
+    pub fn get_config(&self) -> &Config {
+        &self.config
+    }
+
     fn remove_excluded(&mut self) {
         let config = &self.config;
         // FIXME: interpret `config.export.exclude` as `Path`s.
