@@ -2,6 +2,7 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
+use std::collections::hash_map::Entry;
 use std::collections::{HashMap, HashSet};
 use std::fs::File;
 use std::io::Read;
@@ -17,7 +18,7 @@ use crate::bindgen::ir::{
     AnnotationSet, Cfg, Constant, Documentation, Enum, Function, GenericParams, ItemMap,
     OpaqueItem, Path, Static, Struct, Type, Typedef, Union,
 };
-use crate::bindgen::utilities::{SynAbiHelpers, SynItemHelpers};
+use crate::bindgen::utilities::{SynAbiHelpers, SynAttributeHelpers, SynItemFnHelpers};
 
 const STD_CRATES: &[&str] = &[
     "std",
@@ -55,7 +56,7 @@ pub fn parse_src(src_file: &FilePath, config: &Config) -> ParseResult {
         version: None,
     };
 
-    context.parse_mod(&pkg_ref, src_file)?;
+    context.parse_mod(&pkg_ref, src_file, 0)?;
     Ok(context.out)
 }
 
@@ -114,6 +115,7 @@ impl<'a> Parser<'a> {
         // If we have a whitelist, check it
         if let Some(ref include) = self.config.parse.include {
             if !include.contains(&pkg_name) {
+                debug!("Excluding crate {}", pkg_name);
                 return false;
             }
         }
@@ -124,6 +126,7 @@ impl<'a> Parser<'a> {
 
     fn parse_crate(&mut self, pkg: &PackageRef) -> Result<(), Error> {
         assert!(self.lib.is_some());
+        debug!("Parsing crate {}", pkg.name);
         self.parsed_crates.insert(pkg.name.clone());
 
         // Check if we should use cargo expand for this crate
@@ -136,7 +139,7 @@ impl<'a> Parser<'a> {
             let crate_src = self.lib.as_ref().unwrap().find_crate_src(pkg);
 
             match crate_src {
-                Some(crate_src) => self.parse_mod(pkg, crate_src.as_path())?,
+                Some(crate_src) => self.parse_mod(pkg, crate_src.as_path(), 0)?,
                 None => {
                     // This should be an error, but is common enough to just elicit a warning
                     warn!(
@@ -176,7 +179,7 @@ impl<'a> Parser<'a> {
             return Ok(());
         }
 
-        let mod_parsed = {
+        let mod_items = {
             if !self.cache_expanded_crate.contains_key(&pkg.name) {
                 let s = self
                     .lib
@@ -200,53 +203,23 @@ impl<'a> Parser<'a> {
             self.cache_expanded_crate.get(&pkg.name).unwrap().clone()
         };
 
-        self.process_expanded_mod(pkg, &mod_parsed)
+        self.process_mod(pkg, None, &mod_items, 0)
     }
 
-    fn process_expanded_mod(&mut self, pkg: &PackageRef, items: &[syn::Item]) -> Result<(), Error> {
-        self.out.load_syn_crate_mod(
-            &self.config,
-            &self.binding_crate_name,
-            &pkg.name,
-            Cfg::join(&self.cfg_stack).as_ref(),
-            items,
-        );
-
-        for item in items {
-            if item.has_test_attr() {
-                continue;
-            }
-            if let syn::Item::Mod(ref item) = *item {
-                let cfg = Cfg::load(&item.attrs);
-                if let Some(ref cfg) = cfg {
-                    self.cfg_stack.push(cfg.clone());
-                }
-
-                if let Some((_, ref inline_items)) = item.content {
-                    self.process_expanded_mod(pkg, inline_items)?;
-                } else {
-                    unreachable!();
-                }
-
-                if cfg.is_some() {
-                    self.cfg_stack.pop();
-                }
-            }
-        }
-
-        Ok(())
-    }
-
-    fn parse_mod(&mut self, pkg: &PackageRef, mod_path: &FilePath) -> Result<(), Error> {
-        let mod_parsed = {
-            let owned_mod_path = mod_path.to_path_buf();
-
-            if !self.cache_src.contains_key(&owned_mod_path) {
+    fn parse_mod(
+        &mut self,
+        pkg: &PackageRef,
+        mod_path: &FilePath,
+        depth: usize,
+    ) -> Result<(), Error> {
+        let mod_items = match self.cache_src.entry(mod_path.to_path_buf()) {
+            Entry::Vacant(vacant_entry) => {
                 let mut s = String::new();
                 let mut f = File::open(mod_path).map_err(|_| Error::ParseCannotOpenFile {
                     crate_name: pkg.name.clone(),
                     src_path: mod_path.to_str().unwrap().to_owned(),
                 })?;
+
                 f.read_to_string(&mut s)
                     .map_err(|_| Error::ParseCannotOpenFile {
                         crate_name: pkg.name.clone(),
@@ -255,28 +228,42 @@ impl<'a> Parser<'a> {
 
                 let i = syn::parse_file(&s).map_err(|x| Error::ParseSyntaxError {
                     crate_name: pkg.name.clone(),
-                    src_path: owned_mod_path.to_string_lossy().into(),
+                    src_path: mod_path.to_string_lossy().into(),
                     error: x,
                 })?;
 
-                self.cache_src.insert(owned_mod_path.clone(), i.items);
+                vacant_entry.insert(i.items).clone()
             }
-
-            self.cache_src.get(&owned_mod_path).unwrap().clone()
+            Entry::Occupied(occupied_entry) => occupied_entry.get().clone(),
         };
 
-        let mod_dir = mod_path.parent().unwrap();
+        // Compute module directory according to Rust 2018 rules
+        let mod_dir_2018;
 
-        self.process_mod(pkg, mod_dir, &mod_parsed)
+        let mod_dir = if depth == 0 || mod_path.ends_with("mod.rs") {
+            mod_path.parent().unwrap()
+        } else {
+            mod_dir_2018 = mod_path
+                .parent()
+                .unwrap()
+                .join(mod_path.file_stem().unwrap());
+            &mod_dir_2018
+        };
+
+        self.process_mod(pkg, Some(&mod_dir), &mod_items, depth)
     }
 
+    /// `mod_dir` is the path to the current directory of the module. It may be
+    /// `None` for pre-expanded modules.
     fn process_mod(
         &mut self,
         pkg: &PackageRef,
-        mod_dir: &FilePath,
+        mod_dir: Option<&FilePath>,
         items: &[syn::Item],
+        depth: usize,
     ) -> Result<(), Error> {
-        self.out.load_syn_crate_mod(
+        // We process the items first then the nested modules.
+        let nested_modules = self.out.load_syn_crate_mod(
             &self.config,
             &self.binding_crate_name,
             &pkg.name,
@@ -284,61 +271,71 @@ impl<'a> Parser<'a> {
             items,
         );
 
-        for item in items {
-            if item.has_test_attr() {
-                continue;
+        for item in nested_modules {
+            let next_mod_name = item.ident.to_string();
+
+            let cfg = Cfg::load(&item.attrs);
+            if let Some(ref cfg) = cfg {
+                self.cfg_stack.push(cfg.clone());
             }
-            if let syn::Item::Mod(ref item) = *item {
-                let next_mod_name = item.ident.to_string();
 
-                let cfg = Cfg::load(&item.attrs);
-                if let Some(ref cfg) = cfg {
-                    self.cfg_stack.push(cfg.clone());
-                }
+            if let Some((_, ref inline_items)) = item.content {
+                let next_mod_dir = mod_dir.map(|dir| dir.join(&next_mod_name));
+                self.process_mod(
+                    pkg,
+                    next_mod_dir.as_ref().map(|p| &**p),
+                    inline_items,
+                    depth,
+                )?;
+            } else if let Some(mod_dir) = mod_dir {
+                let next_mod_path1 = mod_dir.join(next_mod_name.clone() + ".rs");
+                let next_mod_path2 = mod_dir.join(next_mod_name.clone()).join("mod.rs");
 
-                if let Some((_, ref inline_items)) = item.content {
-                    self.process_mod(pkg, &mod_dir.join(&next_mod_name), inline_items)?;
+                if next_mod_path1.exists() {
+                    self.parse_mod(pkg, next_mod_path1.as_path(), depth + 1)?;
+                } else if next_mod_path2.exists() {
+                    self.parse_mod(pkg, next_mod_path2.as_path(), depth + 1)?;
                 } else {
-                    let next_mod_path1 = mod_dir.join(next_mod_name.clone() + ".rs");
-                    let next_mod_path2 = mod_dir.join(next_mod_name.clone()).join("mod.rs");
-
-                    if next_mod_path1.exists() {
-                        self.parse_mod(pkg, next_mod_path1.as_path())?;
-                    } else if next_mod_path2.exists() {
-                        self.parse_mod(pkg, next_mod_path2.as_path())?;
-                    } else {
-                        // Last chance to find a module path
-                        let mut path_attr_found = false;
-                        for attr in &item.attrs {
-                            match attr.parse_meta() {
-                                Ok(syn::Meta::NameValue(syn::MetaNameValue {
-                                    path, lit, ..
-                                })) => match lit {
+                    // Last chance to find a module path
+                    let mut path_attr_found = false;
+                    for attr in &item.attrs {
+                        match attr.parse_meta() {
+                            Ok(syn::Meta::NameValue(syn::MetaNameValue { path, lit, .. })) => {
+                                match lit {
                                     syn::Lit::Str(ref path_lit) if path.is_ident("path") => {
                                         path_attr_found = true;
-                                        self.parse_mod(pkg, &mod_dir.join(path_lit.value()))?;
+                                        self.parse_mod(
+                                            pkg,
+                                            &mod_dir.join(path_lit.value()),
+                                            depth + 1,
+                                        )?;
                                         break;
                                     }
                                     _ => (),
-                                },
-                                _ => (),
+                                }
                             }
-                        }
-
-                        // This should be an error, but it's common enough to
-                        // just elicit a warning
-                        if !path_attr_found {
-                            warn!(
-                                "Parsing crate `{}`: can't find mod {}`.",
-                                pkg.name, next_mod_name
-                            );
+                            _ => (),
                         }
                     }
-                }
 
-                if cfg.is_some() {
-                    self.cfg_stack.pop();
+                    // This should be an error, but it's common enough to
+                    // just elicit a warning
+                    if !path_attr_found {
+                        warn!(
+                            "Parsing crate `{}`: can't find mod {}`.",
+                            pkg.name, next_mod_name
+                        );
+                    }
                 }
+            } else {
+                warn!(
+                    "Parsing expanded crate `{}`: can't find mod {}`.",
+                    pkg.name, next_mod_name
+                );
+            }
+
+            if cfg.is_some() {
+                self.cfg_stack.pop();
             }
         }
 
@@ -387,13 +384,14 @@ impl Parse {
 
         add_opaque("String", vec![]);
         add_opaque("Box", vec!["T"]);
+        add_opaque("RefCell", vec!["T"]);
         add_opaque("Rc", vec!["T"]);
         add_opaque("Arc", vec!["T"]);
         add_opaque("Result", vec!["T", "E"]);
         add_opaque("Option", vec!["T"]);
         add_opaque("NonNull", vec!["T"]);
         add_opaque("Vec", vec!["T"]);
-        add_opaque("HashMap", vec!["K", "V"]);
+        add_opaque("HashMap", vec!["K", "V", "Hasher"]);
         add_opaque("BTreeMap", vec!["K", "V"]);
         add_opaque("HashSet", vec!["T"]);
         add_opaque("BTreeSet", vec!["T"]);
@@ -412,18 +410,19 @@ impl Parse {
         self.functions.extend_from_slice(&other.functions);
     }
 
-    pub fn load_syn_crate_mod(
+    fn load_syn_crate_mod<'a>(
         &mut self,
         config: &Config,
         binding_crate_name: &str,
         crate_name: &str,
         mod_cfg: Option<&Cfg>,
-        items: &[syn::Item],
-    ) {
+        items: &'a [syn::Item],
+    ) -> Vec<&'a syn::ItemMod> {
         let mut impls_with_assoc_consts = Vec::new();
+        let mut nested_modules = Vec::new();
 
         for item in items {
-            if item.has_test_attr() {
+            if item.should_skip_parsing() {
                 continue;
             }
             match item {
@@ -452,7 +451,7 @@ impl Parse {
                     self.load_syn_union(config, crate_name, mod_cfg, item);
                 }
                 syn::Item::Enum(ref item) => {
-                    self.load_syn_enum(crate_name, mod_cfg, item);
+                    self.load_syn_enum(config, crate_name, mod_cfg, item);
                 }
                 syn::Item::Type(ref item) => {
                     self.load_syn_ty(crate_name, mod_cfg, item);
@@ -465,10 +464,31 @@ impl Parse {
                     if has_assoc_const {
                         impls_with_assoc_consts.push(item_impl);
                     }
+
+                    if let syn::Type::Path(ref path) = *item_impl.self_ty {
+                        if let Some(type_name) = path.path.get_ident() {
+                            for method in item_impl.items.iter().filter_map(|item| match item {
+                                syn::ImplItem::Method(method) => Some(method),
+                                _ => None,
+                            }) {
+                                self.load_syn_method(
+                                    config,
+                                    binding_crate_name,
+                                    crate_name,
+                                    mod_cfg,
+                                    &Path::new(type_name.to_string()),
+                                    method,
+                                )
+                            }
+                        }
+                    }
                 }
                 syn::Item::Macro(ref item) => {
                     self.load_builtin_macro(config, crate_name, mod_cfg, item)
                 }
+                syn::Item::Mod(ref item) => {
+                    nested_modules.push(item);
+                }
                 _ => {}
             }
         }
@@ -476,6 +496,8 @@ impl Parse {
         for item_impl in impls_with_assoc_consts {
             self.load_syn_assoc_consts_from_impl(crate_name, mod_cfg, item_impl)
         }
+
+        nested_modules
     }
 
     fn load_syn_assoc_consts_from_impl(
@@ -523,7 +545,7 @@ impl Parse {
                     return;
                 }
                 let path = Path::new(function.sig.ident.to_string());
-                match Function::load(path, &function.sig, true, &function.attrs, mod_cfg) {
+                match Function::load(path, None, &function.sig, true, &function.attrs, mod_cfg) {
                     Ok(func) => {
                         info!("Take {}::{}.", crate_name, &function.sig.ident);
 
@@ -540,6 +562,29 @@ impl Parse {
         }
     }
 
+    /// Loads a `fn` declaration inside an `impl` block, if the type is a simple identifier
+    fn load_syn_method(
+        &mut self,
+        config: &Config,
+        binding_crate_name: &str,
+        crate_name: &str,
+        mod_cfg: Option<&Cfg>,
+        self_type: &Path,
+        item: &syn::ImplItemMethod,
+    ) {
+        self.load_fn_declaration(
+            config,
+            binding_crate_name,
+            crate_name,
+            mod_cfg,
+            item,
+            Some(self_type),
+            &item.sig,
+            &item.vis,
+            &item.attrs,
+        )
+    }
+
     /// Loads a `fn` declaration
     fn load_syn_fn(
         &mut self,
@@ -548,6 +593,31 @@ impl Parse {
         crate_name: &str,
         mod_cfg: Option<&Cfg>,
         item: &syn::ItemFn,
+    ) {
+        self.load_fn_declaration(
+            config,
+            binding_crate_name,
+            crate_name,
+            mod_cfg,
+            item,
+            None,
+            &item.sig,
+            &item.vis,
+            &item.attrs,
+        );
+    }
+
+    fn load_fn_declaration(
+        &mut self,
+        config: &Config,
+        binding_crate_name: &str,
+        crate_name: &str,
+        mod_cfg: Option<&Cfg>,
+        named_symbol: &dyn SynItemFnHelpers,
+        self_type: Option<&Path>,
+        sig: &syn::Signature,
+        vis: &syn::Visibility,
+        attrs: &[syn::Attribute],
     ) {
         if !config
             .parse
@@ -555,47 +625,71 @@ impl Parse {
         {
             info!(
                 "Skip {}::{} - (fn's outside of the binding crate are not used).",
-                crate_name, &item.sig.ident
+                crate_name, &sig.ident
             );
             return;
         }
 
-        if let syn::Visibility::Public(_) = item.vis {
-            if item.is_no_mangle() && (item.sig.abi.is_omitted() || item.sig.abi.is_c()) {
-                let path = Path::new(item.sig.ident.to_string());
-                match Function::load(path, &item.sig, false, &item.attrs, mod_cfg) {
-                    Ok(func) => {
-                        info!("Take {}::{}.", crate_name, &item.sig.ident);
+        let loggable_item_name = || {
+            let mut items = vec![];
+            items.push(crate_name.to_owned());
+            if let Some(ref self_type) = self_type {
+                items.push(self_type.to_string());
+            }
+            items.push(sig.ident.to_string());
+            items.join("::")
+        };
 
-                        self.functions.push(func);
-                    }
-                    Err(msg) => {
-                        error!(
-                            "Cannot use fn {}::{} ({}).",
-                            crate_name, &item.sig.ident, msg
-                        );
+        let is_extern_c = sig.abi.is_omitted() || sig.abi.is_c();
+        let exported_name = named_symbol.exported_name();
+
+        if let syn::Visibility::Public(_) = vis {
+            match (is_extern_c, exported_name) {
+                (true, Some(exported_name)) => {
+                    let path = Path::new(exported_name);
+                    match Function::load(path, self_type, &sig, false, &attrs, mod_cfg) {
+                        Ok(func) => {
+                            info!("Take {}.", loggable_item_name());
+                            self.functions.push(func);
+                        }
+                        Err(msg) => {
+                            error!("Cannot use fn {} ({}).", loggable_item_name(), msg);
+                        }
                     }
                 }
-                return;
+                (true, None) => {
+                    warn!(
+                        "Skipping {} - (not `no_mangle`, and has no `export_name` attribute)",
+                        loggable_item_name()
+                    );
+                }
+                (false, Some(_exported_name)) => {
+                    warn!("Skipping {} - (not `extern \"C\"`", loggable_item_name());
+                }
+                (false, None) => {}
             }
-        }
-
-        // TODO
-        if let syn::Visibility::Public(_) = item.vis {
         } else {
-            warn!("Skip {}::{} - (not `pub`).", crate_name, &item.sig.ident);
-        }
-        if (item.sig.abi.is_omitted() || item.sig.abi.is_c()) && !item.is_no_mangle() {
-            warn!(
-                "Skip {}::{} - (`extern` but not `no_mangle`).",
-                crate_name, &item.sig.ident
-            );
-        }
-        if item.sig.abi.is_some() && !(item.sig.abi.is_omitted() || item.sig.abi.is_c()) {
-            warn!(
-                "Skip {}::{} - (non `extern \"C\"`).",
-                crate_name, &item.sig.ident
-            );
+            match (is_extern_c, exported_name) {
+                (true, Some(..)) => {
+                    warn!(
+                        "Skipping {} - (not `pub` but is `extern \"C\"` and `no_mangle`)",
+                        loggable_item_name()
+                    );
+                }
+                (true, None) => {
+                    warn!(
+                        "Skipping {} - (not `pub` but is `extern \"C\"`)",
+                        loggable_item_name()
+                    );
+                }
+                (false, Some(..)) => {
+                    warn!(
+                        "Skipping {} - (not `pub` but is `no_mangle`)",
+                        loggable_item_name()
+                    );
+                }
+                (false, None) => {}
+            }
         }
     }
 
@@ -616,11 +710,22 @@ impl Parse {
                 return;
             }
         };
-        if ty.is_none() {
-            return;
-        }
 
-        let impl_path = ty.unwrap().get_root_path().unwrap();
+        let ty = match ty {
+            Some(ty) => ty,
+            None => return,
+        };
+
+        let impl_path = match ty.get_root_path() {
+            Some(p) => p,
+            None => {
+                warn!(
+                    "Couldn't find path for {:?}, skipping associated constants",
+                    ty
+                );
+                return;
+            }
+        };
 
         for item in items.into_iter() {
             if let syn::Visibility::Public(_) = item.vis {
@@ -796,7 +901,13 @@ impl Parse {
     }
 
     /// Loads a `enum` declaration
-    fn load_syn_enum(&mut self, crate_name: &str, mod_cfg: Option<&Cfg>, item: &syn::ItemEnum) {
+    fn load_syn_enum(
+        &mut self,
+        config: &Config,
+        crate_name: &str,
+        mod_cfg: Option<&Cfg>,
+        item: &syn::ItemEnum,
+    ) {
         if item.generics.lifetimes().count() > 0 {
             info!(
                 "Skip {}::{} - (has generics or lifetimes or where bounds).",
@@ -805,7 +916,7 @@ impl Parse {
             return;
         }
 
-        match Enum::load(item, mod_cfg) {
+        match Enum::load(item, mod_cfg, config) {
             Ok(en) => {
                 info!("Take {}::{}.", crate_name, &item.ident);
                 self.enums.try_insert(en);
